import numpy as np

from autocti.charge_injection import ci_frame, ci_pattern
from autocti.data import mask as msk
from test.unit.mock.mock import MockCIGeometry


class TestBinArrayAcrossSerial:

    def test__3x3_array__all_1s__bin_gives_a_1d_array_of_3_1s(self):
        image = np.ones((3, 3))

        binned_array = ci_frame.bin_array_across_serial(image)

        assert (binned_array == np.array([1.0, 1.0, 1.0])).all()

    def test__4x3_array__all_1s__bin_gives_a_1d_array_of_4_1s(self):
        image = np.ones((4, 3))

        binned_array = ci_frame.bin_array_across_serial(image)

        assert (binned_array == np.array([1.0, 1.0, 1.0, 1.0])).all()

    def test__3x4_array__all_1s__bin_gives_a_1d_array_of_3_1s(self):
        image = np.ones((3, 4))

        binned_array = ci_frame.bin_array_across_serial(image)

        assert (binned_array == np.array([1.0, 1.0, 1.0])).all()

    def test__3x3_array__different_values__bin_gives_a_1d_array_of_each_mean(self):
        image = np.array([[1.0, 2.0, 3.0],
                          [6.0, 6.0, 6.0],
                          [9.0, 9.0, 9.0]])

        binned_array = ci_frame.bin_array_across_serial(image)

        assert (binned_array == np.array([2.0, 6.0, 9.0])).all()

    def test__3x3_array__same_as_above_but_including_mask(self):
        image = np.array([[1.0, 2.0, 3.0],
                          [6.0, 6.0, 6.0],
                          [9.0, 9.0, 9.0]])

        mask = np.ma.array([[False, False, True],
                            [False, False, False],
                            [False, False, False]])

        binned_array = ci_frame.bin_array_across_serial(image, mask)

        assert (binned_array == np.array([1.5, 6.0, 9.0])).all()

    def test__3x3_array__same_as_above_but_an_entire_row_is_masked(self):
        image = np.array([[1.0, 2.0, 3.0],
                          [6.0, 6.0, 6.0],
                          [9.0, 9.0, 9.0]])

        mask = np.ma.array([[False, False, True],
                            [True, True, True],
                            [False, False, False]])

        binned_array = ci_frame.bin_array_across_serial(image, mask)

        assert (binned_array == np.array([1.5, np.inf, 9.0])).all()


class TestBinArrayAcrossParallel:

    def test__3x3_array__all_1s__bin_gives_a_1d_array_of_3_1s(self):
        image = np.ones((3, 3))

        binned_array = ci_frame.bin_array_across_parallel(image)

        assert (binned_array == np.array([1.0, 1.0, 1.0])).all()

    def test__4x3_array__all_1s__bin_gives_a_1d_array_of_3_1s(self):
        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

        image = np.ones((4, 3))

        binned_array = ci_frame.bin_array_across_parallel(image)

        assert (binned_array == np.array([1.0, 1.0, 1.0])).all()

    def test__3x4_array__all_1s__bin_gives_a_1d_array_of_4_1s(self):
        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

        image = np.ones((3, 4))

        binned_array = ci_frame.bin_array_across_parallel(image)

        assert (binned_array == np.array([1.0, 1.0, 1.0, 1.0])).all()

    def test__3x3_array__different_values__bin_gives_a_1d_array_of_each_mean(self):
        image = np.array([[1.0, 6.0, 9.0],
                          [2.0, 6.0, 9.0],
                          [3.0, 6.0, 9.0]])

        binned_array = ci_frame.bin_array_across_parallel(image)

        assert (binned_array == np.array([2.0, 6.0, 9.0])).all()

    def test__3x3_array__same_as_above_but_with_mask(self):
        image = np.array([[1.0, 6.0, 9.0],
                          [2.0, 6.0, 9.0],
                          [3.0, 6.0, 9.0]])

        mask = np.ma.array([[False, False, False],
                            [False, False, False],
                            [True, False, False]])

        binned_array = ci_frame.bin_array_across_parallel(image, mask)

        assert (binned_array == np.array([1.5, 6.0, 9.0])).all()

    def test__3x3_array__same_as_above_but_with_entire_column_masked(self):
        image = np.array([[1.0, 6.0, 9.0],
                          [2.0, 6.0, 9.0],
                          [3.0, 6.0, 9.0]])

        mask = np.ma.array([[False, True, False],
                            [False, True, False],
                            [True, True, False]])

        binned_array = ci_frame.bin_array_across_parallel(image, mask)

        assert (binned_array == np.array([1.5, np.inf, 9.0])).all()


class TestChInj(object):

    class TestCiRegionArrayFromFrame:

        def test__1_ci_region__extracted_correctly(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 3)])
            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.ci_regions_from_array(image, )

            assert (new_frame == np.array([[0.0, 1.0, 2.0],
                                           [3.0, 4.0, 5.0],
                                           [6.0, 7.0, 8.0],
                                           [0.0, 0.0, 0.0]])).all()

        def test__2_ci_regions__extracted_correctly(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 1, 1, 2), (2, 3, 1, 3)])
            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.ci_regions_from_array(image, )

            assert (new_frame == np.array([[0.0, 1.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 7.0, 8.0],
                                           [0.0, 0.0, 0.0]])).all()

    class TestCIParallelNonRegionArrayFromFrame:

        def test__1_ci_region__pre_scan_and_overscan_in_corner__extracts_everything_outside_region_correctly(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0]])

            frame = ci_frame.ChInj(
                frame_geometry=MockCIGeometry(serial_prescan=(0, 1, 0, 1), serial_overscan=(0, 1, 0, 1)),
                ci_pattern=pattern)

            new_frame = frame.parallel_non_ci_regions_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [9.0, 10.0, 11.0]])).all()

        def test__2_ci_regions__pre_scan_and_overscan_in_corner__extracts_everything_outside_region_correctly(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 1, 0, 3), (3, 4, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0]])

            frame = ci_frame.ChInj(
                frame_geometry=MockCIGeometry(serial_prescan=(0, 1, 0, 1), serial_overscan=(0, 1, 0, 1)),
                ci_pattern=pattern)

            new_frame = frame.parallel_non_ci_regions_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 0.0],
                                           [3.0, 4.0, 5.0],
                                           [6.0, 7.0, 8.0],
                                           [0.0, 0.0, 0.0],
                                           [12.0, 13.0, 14.0]])).all()

        def test__2_ci_regions__serial_prescan_overlaps_an_extraction__extraction_goes_to_0(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 1, 0, 3), (3, 4, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0]])

            frame = ci_frame.ChInj(frame_geometry=MockCIGeometry(serial_prescan=(1, 2, 0, 2),
                                                                   serial_overscan=(0, 1, 0, 1)),
                                     ci_pattern=pattern)

            new_frame = frame.parallel_non_ci_regions_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 0.0],
                                           [0.0, 0.0, 5.0],
                                           [6.0, 7.0, 8.0],
                                           [0.0, 0.0, 0.0],
                                           [12.0, 13.0, 14.0]])).all()

        def test__2_ci_regions__serial_overscan_overlaps_an_extraction__extraction_goes_to_0(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 1, 0, 3), (3, 4, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0]])

            frame = ci_frame.ChInj(frame_geometry=MockCIGeometry(serial_prescan=(0, 1, 0, 1),
                                                                   serial_overscan=(1, 2, 1, 3)),
                                     ci_pattern=pattern)

            new_frame = frame.parallel_non_ci_regions_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 0.0],
                                           [3.0, 0.0, 0.0],
                                           [6.0, 7.0, 8.0],
                                           [0.0, 0.0, 0.0],
                                           [12.0, 13.0, 14.0]])).all()

    class TestParallelEdgesAndTrailsArrayFromFrame:

        def test__front_edge_only__1_row__new_frame_is_just_that_edge(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.parallel_edges_and_trails_frame_from_frame(image, front_edge_rows=(0, 1))

            assert (new_frame == np.array([[0.0, 1.0, 2.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])).all()

        def test__front_edge_only__2_rows__new_frame_is_just_that_edge(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.parallel_edges_and_trails_frame_from_frame(image, front_edge_rows=(0, 2))

            assert (new_frame == np.array([[0.0, 1.0, 2.0],
                                           [3.0, 4.0, 5.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0]])).all()

        def test__trails_only__1_row__new_frame_is_just_that_trail(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 4, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.parallel_edges_and_trails_frame_from_frame(image, trails_rows=(0, 1))

            assert (new_frame == np.array([[0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [12.0, 13.0, 14.0]])).all()

        def test__trails_only__2_rows__new_frame_is_the_trails(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 4, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0],
                              [15.0, 16.0, 17.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.parallel_edges_and_trails_frame_from_frame(image, trails_rows=(0, 2))

            assert (new_frame == np.array([[0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [12.0, 13.0, 14.0],
                                           [15.0, 16.0, 17.0]])).all()

        def test__front_edge_and_trails__2_rows_of_each__new_frame_is_edge_and_trail(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 4, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0],
                              [15.0, 16.0, 17.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.parallel_edges_and_trails_frame_from_frame(image, front_edge_rows=(0, 2),
                                                                         trails_rows=(0, 2))

            assert (new_frame == np.array([[0.0, 1.0, 2.0],
                                           [3.0, 4.0, 5.0],
                                           [0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0],
                                           [12.0, 13.0, 14.0],
                                           [15.0, 16.0, 17.0]])).all()

        def test__front_edge_and_trails__2_regions__1_row_of_each__new_frame_is_edge_and_trail(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 1, 0, 3), (3, 4, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0],
                              [15.0, 16.0, 17.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.parallel_edges_and_trails_frame_from_frame(image, front_edge_rows=(0, 1),
                                                                         trails_rows=(0, 1))

            assert (new_frame == np.array([[0.0, 1.0, 2.0],
                                           [3.0, 4.0, 5.0],
                                           [0.0, 0.0, 0.0],
                                           [9.0, 10.0, 11.0],
                                           [12.0, 13.0, 14.0],
                                           [0.0, 0.0, 0.0]])).all()

    class TesParallelCalibrationSectionFromFrame:

        def test__geometry_left__columns_0_to_1__extracts_1_column_left_hand_side_of_array(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 3, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            extracted_side = frame.parallel_calibration_section_for_columns(image, columns=(0, 1))

            assert (extracted_side == np.array([[0.0],
                                                [0.0],
                                                [0.0],
                                                [0.0],
                                                [0.0]])).all()

        def test__geometry_bottom__columns_1_to_3__extracts_2_columns_middle_and_right_of_array(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 5, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            extracted_side = frame.parallel_calibration_section_for_columns(image, columns=(1, 3))

            assert (extracted_side == np.array([[1.0, 2.0],
                                                [1.0, 2.0],
                                                [1.0, 2.0],
                                                [1.0, 2.0],
                                                [1.0, 2.0]])).all()

        def test__geometry_right__columns_1_to_3__extracts_2_columns_middle_and_left_of_array(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 5, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0],
                              [0.0, 1.0, 2.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

            extracted_side = frame.parallel_calibration_section_for_columns(image, columns=(1, 3))

            assert (extracted_side == np.array([[0.0, 1.0],
                                                [0.0, 1.0],
                                                [0.0, 1.0],
                                                [0.0, 1.0],
                                                [0.0, 1.0]])).all()

    class TestKeepSerialEdgesAndTrailsArrayFromFrame:

        def test__front_edge_only__1_column__new_frame_is_just_that_edge(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0, 7.0],
                              [8.0, 9.0, 10.0, 11.0]])
            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.serial_edges_and_trails_frame_from_frame(image, front_edge_columns=(0, 1))

            assert (new_frame == np.array([[0.0, 0.0, 0.0, 0.0],
                                           [4.0, 0.0, 0.0, 0.0],
                                           [8.0, 0.0, 0.0, 0.0]])).all()

        def test__front_edge_only__2_columns__new_frame_is_just_that_edge(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 3)])

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0, 7.0],
                              [8.0, 9.0, 10.0, 11.0]])
            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.serial_edges_and_trails_frame_from_frame(image, front_edge_columns=(0, 2))

            assert (new_frame == np.array([[0.0, 1.0, 0.0, 0.0],
                                           [4.0, 5.0, 0.0, 0.0],
                                           [8.0, 9.0, 0.0, 0.0]])).all()

        def test__trails_only__1_column__new_frame_is_just_that_trail(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 2)])

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0, 7.0],
                              [8.0, 9.0, 10.0, 11.0]])
            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.serial_edges_and_trails_frame_from_frame(image, trails_columns=(0, 1))

            assert (new_frame == np.array([[0.0, 0.0, 2.0, 0.0],
                                           [0.0, 0.0, 6.0, 0.0],
                                           [0.0, 0.0, 10.0, 0.0]])).all()

        def test__trails_only__2_columns__new_frame_is_the_trails(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 2)])

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0, 7.0],
                              [8.0, 9.0, 10.0, 11.0]])
            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.serial_edges_and_trails_frame_from_frame(image, trails_columns=(0, 2))

            assert (new_frame == np.array([[0.0, 0.0, 2.0, 3.0],
                                           [0.0, 0.0, 6.0, 7.0],
                                           [0.0, 0.0, 10.0, 11.0]])).all()

        def test__front_edge_and_trails__2_columns_of_each__new_frame_is_edge_and_trail(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 2)])

            image = np.array([[0.0, 1.0, 1.1, 2.0, 3.0],
                              [4.0, 5.0, 1.1, 6.0, 7.0],
                              [8.0, 9.0, 1.1, 10.0, 11.0]])
            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.serial_edges_and_trails_frame_from_frame(image, front_edge_columns=(0, 1),
                                                                       trails_columns=(0, 2))

            assert (new_frame == np.array([[0.0, 0.0, 1.1, 2.0, 0.0],
                                           [4.0, 0.0, 1.1, 6.0, 0.0],
                                           [8.0, 0.0, 1.1, 10.0, 0.0]])).all()

        def test__front_edge_and_trails__2_regions_1_column_of_each__new_frame_is_edge_and_trail(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 3, 0, 1), (0, 3, 3, 4)])

            image = np.array([[0.0, 1.0, 1.1, 2.0, 3.0],
                              [4.0, 5.0, 1.1, 6.0, 7.0],
                              [8.0, 9.0, 1.1, 10.0, 11.0]])
            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(),
                                   ci_pattern=pattern)

            new_frame = frame.serial_edges_and_trails_frame_from_frame(image, front_edge_columns=(0, 1),
                                                                       trails_columns=(0, 1))

            assert (new_frame == np.array([[0.0, 1.0, 0.0, 2.0, 3.0],
                                           [4.0, 5.0, 0.0, 6.0, 7.0],
                                           [8.0, 9.0, 0.0, 10.0, 11.0]])).all()

    class TestSerialAllTrailsArrayFromFrame:

        def test__left_quadrant__1_ci_region__1_serial_trail__extracts_all_trails(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 4, 0, 2)])

            frame_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            frame_geometry.serial_overscan = ci_frame.Region((0, 4, 2, 3))

            image = np.array([[0.0, 1.0, 2.0],
                              [3.0, 4.0, 5.0],
                              [6.0, 7.0, 8.0],
                              [9.0, 10.0, 11.0]])

            frame = ci_frame.ChInj(frame_geometry=frame_geometry,
                                     ci_pattern=pattern)

            new_frame = frame.serial_all_trails_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 2.0],
                                           [0.0, 0.0, 5.0],
                                           [0.0, 0.0, 8.0],
                                           [0.0, 0.0, 11.0]])).all()

        def test__left_quadrant__1_ci_region__2_serial_trail__extracts_all_trails(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 4, 0, 2)])

            frame_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            frame_geometry.serial_overscan = ci_frame.Region((0, 4, 2, 4))

            image = np.array([[0.0, 1.0, 2.0, 0.5],
                              [3.0, 4.0, 5.0, 0.5],
                              [6.0, 7.0, 8.0, 0.5],
                              [9.0, 10.0, 11.0, 0.5]])

            frame = ci_frame.ChInj(frame_geometry=frame_geometry,
                                     ci_pattern=pattern)

            new_frame = frame.serial_all_trails_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 2.0, 0.5],
                                           [0.0, 0.0, 5.0, 0.5],
                                           [0.0, 0.0, 8.0, 0.5],
                                           [0.0, 0.0, 11.0, 0.5]])).all()

        def test__left_quadrant__2_ci_regions__2_serial_trail__extracts_all_trails(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 1, 0, 2), (2, 3, 0, 2)])

            frame_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            frame_geometry.serial_overscan = ci_frame.Region((0, 4, 2, 4))

            image = np.array([[0.0, 1.0, 2.0, 0.5],
                              [3.0, 4.0, 5.0, 0.5],
                              [6.0, 7.0, 8.0, 0.5],
                              [9.0, 10.0, 11.0, 0.5]])

            frame = ci_frame.ChInj(frame_geometry=frame_geometry,
                                     ci_pattern=pattern)

            new_frame = frame.serial_all_trails_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 2.0, 0.5],
                                           [0.0, 0.0, 0.0, 0.0],
                                           [0.0, 0.0, 8.0, 0.5],
                                           [0.0, 0.0, 0.0, 0.0]])).all()

        def test__same_as_above_but_right_quadrant__flips_trails_side(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(0, 1, 2, 4), (2, 3, 2, 4)])

            frame_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            frame_geometry.serial_overscan = ci_frame.Region((0, 4, 0, 2))

            image = np.array([[0.0, 1.0, 2.0, 0.5],
                              [3.0, 4.0, 5.0, 0.5],
                              [6.0, 7.0, 8.0, 0.5],
                              [9.0, 10.0, 11.0, 0.5]])

            frame = ci_frame.ChInj(frame_geometry=frame_geometry,
                                     ci_pattern=pattern)

            new_frame = frame.serial_all_trails_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 1.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0, 0.0],
                                           [6.0, 7.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0, 0.0]])).all()

    class TestSerialOverScanNonTrailsFromFrame:

        def test__left_quadrant__1_ci_region__serial_trails_go_over_2_right_hand_columns__2_pixels_above_kept(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(1, 3, 1, 2)])

            frame_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            frame_geometry.serial_prescan = ci_frame.Region((0, 3, 0, 1))
            frame_geometry.serial_overscan = ci_frame.Region((0, 3, 2, 4))

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0, 7.0],
                              [8.0, 9.0, 10.0, 11.0]])

            frame = ci_frame.ChInj(frame_geometry=frame_geometry,
                                     ci_pattern=pattern)

            new_frame = frame.serial_overscan_above_trails_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 2.0, 3.0],
                                           [0.0, 0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0, 0.0]])).all()

        def test__left_quadrant__1_ci_region__serial_trails_go_over_1_right_hand_column__1_pixel_above_kept(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(1, 3, 1, 3)])

            frame_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            frame_geometry.serial_prescan = ci_frame.Region((0, 3, 0, 1))
            frame_geometry.serial_overscan = ci_frame.Region((0, 3, 3, 4))

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0, 7.0],
                              [8.0, 9.0, 10.0, 11.0]])

            frame = ci_frame.ChInj(frame_geometry=frame_geometry,
                                     ci_pattern=pattern)

            new_frame = frame.serial_overscan_above_trails_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 0.0, 3.0],
                                           [0.0, 0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0, 0.0]])).all()

        def test__left_quadrant__2_ci_regions__serial_trails_go_over_1_right_hand_column__1_pixel_above_each_kept(self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(1, 2, 1, 3), (3, 4, 1, 3)])

            frame_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            frame_geometry.serial_prescan = ci_frame.Region((0, 5, 0, 1))
            frame_geometry.serial_overscan = ci_frame.Region((0, 5, 3, 4))

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0, 7.0],
                              [8.0, 9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0, 15.0],
                              [16.0, 17.0, 18.0, 19.0]])

            frame = ci_frame.ChInj(frame_geometry=frame_geometry,
                                     ci_pattern=pattern)

            new_frame = frame.serial_overscan_above_trails_frame_from_frame(image, )

            assert (new_frame == np.array([[0.0, 0.0, 0.0, 3.0],
                                           [0.0, 0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0, 11.0],
                                           [0.0, 0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0, 19.0]])).all()

        def test__right_quadrant__2_ci_regions__serial_trails_go_over_1_right_hand_column__1_pixel_above_each_kept(
                self):
            pattern = ci_pattern.CIPattern(normalization=10.0, regions=[(1, 2, 1, 3), (3, 4, 1, 3)])

            frame_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            frame_geometry.serial_prescan = ci_frame.Region((0, 5, 3, 4))
            frame_geometry.serial_overscan = ci_frame.Region((0, 5, 0, 1))

            image = np.array([[0.5, 1.0, 2.0, 3.0],
                              [4.0, 5.0, 6.0, 7.0],
                              [8.0, 9.0, 10.0, 11.0],
                              [12.0, 13.0, 14.0, 15.0],
                              [16.0, 17.0, 18.0, 19.0]])

            frame = ci_frame.ChInj(frame_geometry=frame_geometry,
                                     ci_pattern=pattern)

            new_frame = frame.serial_overscan_above_trails_frame_from_frame(image, )

            assert (new_frame == np.array([[0.5, 0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0, 0.0],
                                           [8.0, 0.0, 0.0, 0.0],
                                           [0.0, 0.0, 0.0, 0.0],
                                           [16.0, 0.0, 0.0, 0.0]])).all()

    class TestSerialCalibrationArrayFromFrame:

        def test__geometry_left__ci_region_across_all_image__column_0__extracts_all_columns(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 0, 5)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            serial_frame = frame.serial_calibration_section_for_rows(image, rows=(0, 3))

            assert (serial_frame[0] == np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                                                 [0.0, 1.0, 2.0, 3.0, 4.0],
                                                 [0.0, 1.0, 2.0, 3.0, 4.0]])).all()

        def test__geometry_left__2_ci_regions__both_extracted(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 1, 1, 4), (2, 3, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 2.0, 2.0],
                              [0.0, 1.0, 3.0, 3.0, 3.0],
                              [0.0, 1.0, 4.0, 4.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            serial_frame = frame.serial_calibration_section_for_rows(image, rows=(0, 1))

            assert (serial_frame == np.array([[0.0, 1.0, 2.0, 2.0, 2.0],
                                              [0.0, 1.0, 4.0, 4.0, 4.0]])).all()

        def test__geometry_right__2_ci_regions__both_extracted(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 1, 1, 4), (2, 3, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 2.0, 2.0],
                              [0.0, 1.0, 3.0, 3.0, 3.0],
                              [0.0, 1.0, 4.0, 4.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

            serial_frame = frame.serial_calibration_section_for_rows(image, rows=(0, 1))

            assert (serial_frame == np.array([[0.0, 1.0, 2.0, 2.0, 2.0],
                                              [0.0, 1.0, 4.0, 4.0, 4.0]])).all()

        def test__geometry_left__rows_cuts_out_bottom_row(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 0, 5)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            serial_frame = frame.serial_calibration_section_for_rows(image, rows=(0, 2))

            assert (serial_frame == np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                                              [0.0, 1.0, 2.0, 3.0, 4.0]])).all()

        def test__extract_two_regions_and_cut_bottom_row_from_each(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 2, 1, 4), (3, 5, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 2.0, 2.0],
                              [0.0, 1.0, 3.0, 3.0, 3.0],
                              [0.0, 1.0, 4.0, 4.0, 4.0],
                              [0.0, 1.0, 5.0, 5.0, 5.0],
                              [0.0, 1.0, 6.0, 6.0, 6.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

            serial_frame = frame.serial_calibration_section_for_rows(image, rows=(0, 1))

            assert (serial_frame == np.array([[0.0, 1.0, 2.0, 2.0, 2.0],
                                              [0.0, 1.0, 5.0, 5.0, 5.0]])).all()

        def test__extract_two_regions_and_cut_top_row_from_each(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 2, 1, 4), (3, 5, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 2.0, 2.0],
                              [0.0, 1.0, 3.0, 3.0, 3.0],
                              [0.0, 1.0, 4.0, 4.0, 4.0],
                              [0.0, 1.0, 5.0, 5.0, 5.0],
                              [0.0, 1.0, 6.0, 6.0, 6.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

            serial_frame = frame.serial_calibration_section_for_rows(image, rows=(1, 2))

            assert (serial_frame == np.array([[0.0, 1.0, 3.0, 3.0, 3.0],
                                              [0.0, 1.0, 6.0, 6.0, 6.0]])).all()

    class TestSerialCalibrationSubArrays:

        def test__geometry_left__ci_region_across_all_image__column_0__extracts_all_columns(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 0, 5)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            serial_region = frame.serial_calibration_sub_arrays_from_frame(array=image)

            assert (serial_region[0] == np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                                                  [0.0, 1.0, 2.0, 3.0, 4.0],
                                                  [0.0, 1.0, 2.0, 3.0, 4.0]])).all()

        def test__geometry_left__ci_region_misses_serial_overscan__column_0__extracts_all_columns(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 0, 4)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            serial_region = frame.serial_calibration_sub_arrays_from_frame(array=image)

            assert (serial_region[0] == np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                                                  [0.0, 1.0, 2.0, 3.0, 4.0],
                                                  [0.0, 1.0, 2.0, 3.0, 4.0]])).all()

        def test__geometry_left__ci_region_also_has_prescan__extracts_all_but_1_column(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            serial_region = frame.serial_calibration_sub_arrays_from_frame(array=image)

            assert (serial_region[0] == np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                                                  [0.0, 1.0, 2.0, 3.0, 4.0],
                                                  [0.0, 1.0, 2.0, 3.0, 4.0]])).all()

        def test__geometry_right__ci_region_has_prescan_and_overscan(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

            serial_region = frame.serial_calibration_sub_arrays_from_frame(array=image)

            assert (serial_region[0] == np.array([[0.0, 1.0, 2.0, 3.0, 4.0],
                                                  [0.0, 1.0, 2.0, 3.0, 4.0],
                                                  [0.0, 1.0, 2.0, 3.0, 4.0]])).all()

        def test__geometry_left__2_ci_regions__both_extracted(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 1, 1, 4), (2, 3, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 2.0, 2.0],
                              [0.0, 1.0, 3.0, 3.0, 3.0],
                              [0.0, 1.0, 4.0, 4.0, 4.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            serial_region = frame.serial_calibration_sub_arrays_from_frame(array=image)

            assert (serial_region[0] == np.array([[0.0, 1.0, 2.0, 2.0, 2.0]])).all()
            assert (serial_region[1] == np.array([[0.0, 1.0, 4.0, 4.0, 4.0]])).all()

    class TestParallelFrontEdgesFromFrame:

        def test__pattern_bottom___extracts_1_front_edge_correctly(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 3)])

            image = np.array([[0.0, 0.0, 0.0],
                              [1.0, 1.0, 1.0],  # <- Front edge according to region and this frame_geometry
                              [2.0, 2.0, 2.0],  # <- Next front edge row.
                              [3.0, 3.0, 3.0],
                              [4.0, 4.0, 4.0],
                              [5.0, 5.0, 5.0],
                              [6.0, 6.0, 6.0],
                              [7.0, 7.0, 7.0],
                              [8.0, 8.0, 8.0],
                              [9.0, 9.0, 9.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            front_edge = frame.parallel_front_edge_arrays_from_frame(image, rows=(0, 1))
            assert (front_edge == np.array([[1.0, 1.0, 1.0]])).all()

            front_edge = frame.parallel_front_edge_arrays_from_frame(image, rows=(1, 2))
            assert (front_edge == np.array([[2.0, 2.0, 2.0]])).all()

            front_edge = frame.parallel_front_edge_arrays_from_frame(image, rows=(2, 3))
            assert (front_edge == np.array([[3.0, 3.0, 3.0]])).all()

        def test__pattern_bottom___extracts_multiple_front_edges_correctly(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 5, 0, 3)])

            image = np.array([[0.0, 0.0, 0.0],
                              [1.0, 1.0, 1.0],  # <- Front edge according to region and this frame_geometry
                              [2.0, 2.0, 2.0],  # <- Next front edge row.
                              [3.0, 3.0, 3.0],
                              [4.0, 4.0, 4.0],
                              [5.0, 5.0, 5.0],
                              [6.0, 6.0, 6.0],
                              [7.0, 7.0, 7.0],
                              [8.0, 8.0, 8.0],
                              [9.0, 9.0, 9.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            front_edge = frame.parallel_front_edge_arrays_from_frame(image, rows=(0, 2))
            assert (front_edge == np.array([[1.0, 1.0, 1.0],
                                            [2.0, 2.0, 2.0]])).all()

            front_edge = frame.parallel_front_edge_arrays_from_frame(image, rows=(1, 4))
            assert (front_edge == np.array([[2.0, 2.0, 2.0],
                                            [3.0, 3.0, 3.0],
                                            [4.0, 4.0, 4.0]])).all()

        def test__pattern_bottom__2_regions__extracts_rows_correctly(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 3), (5, 8, 0, 3)])

            image = np.array([[0.0, 0.0, 0.0],
                              [1.0, 1.0, 1.0],  # <- 1st Front edge according to region and this frame_geometry
                              [2.0, 2.0, 2.0],  # <- Next front edge row.
                              [3.0, 3.0, 3.0],
                              [4.0, 4.0, 4.0],
                              [5.0, 5.0, 5.0],  # <- 2nd Front edge according to region and this frame_geometry
                              [6.0, 6.0, 6.0],  # <- Next front edge row.
                              [7.0, 7.0, 7.0],
                              [8.0, 8.0, 8.0],
                              [9.0, 9.0, 9.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            front_edges = frame.parallel_front_edge_arrays_from_frame(image, rows=(0, 1))
            assert (front_edges[0] == np.array([[1.0, 1.0, 1.0]])).all()
            assert (front_edges[1] == np.array([[5.0, 5.0, 5.0]])).all()

            front_edges = frame.parallel_front_edge_arrays_from_frame(image, rows=(1, 2))
            assert (front_edges[0] == np.array([[2.0, 2.0, 2.0]])).all()
            assert (front_edges[1] == np.array([[6.0, 6.0, 6.0]])).all()

            front_edges = frame.parallel_front_edge_arrays_from_frame(image, rows=(2, 3))
            assert (front_edges[0] == np.array([[3.0, 3.0, 3.0]])).all()
            assert (front_edges[1] == np.array([[7.0, 7.0, 7.0]])).all()

            front_edges = frame.parallel_front_edge_arrays_from_frame(image, rows=(0, 3))
            assert (front_edges[0] == np.array([[1.0, 1.0, 1.0],
                                                [2.0, 2.0, 2.0],
                                                [3.0, 3.0, 3.0]])).all()
            assert (front_edges[1] == np.array([[5.0, 5.0, 5.0],
                                                [6.0, 6.0, 6.0],
                                                [7.0, 7.0, 7.0]])).all()

        def test__pattern_top__does_all_the_above_correctly(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 5, 0, 3), (6, 9, 0, 3)])

            image = np.array([[0.0, 0.0, 0.0],
                              [1.0, 1.0, 1.0],
                              [2.0, 2.0, 2.0],
                              [3.0, 3.0, 3.0],  # <- Next front edge row.
                              [4.0, 4.0, 4.0],  # <- 1st Front edge according to region and this frame_geometry
                              [5.0, 5.0, 5.0],
                              [6.0, 6.0, 6.0],
                              [7.0, 7.0, 7.0],  # <- Next front edge row.
                              [8.0, 8.0, 8.0],  # <- 2nd Front edge according to region and this frame_geometry
                              [9.0, 9.0, 9.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), ci_pattern=pattern)

            front_edges = frame.parallel_front_edge_arrays_from_frame(image, rows=(0, 1))
            assert (front_edges[0] == np.array([[4.0, 4.0, 4.0]])).all()
            assert (front_edges[1] == np.array([[8.0, 8.0, 8.0]])).all()

            front_edges = frame.parallel_front_edge_arrays_from_frame(image, rows=(1, 2))
            assert (front_edges[0] == np.array([[3.0, 3.0, 3.0]])).all()
            assert (front_edges[1] == np.array([[7.0, 7.0, 7.0]])).all()

            front_edges = frame.parallel_front_edge_arrays_from_frame(image, rows=(2, 3))
            assert (front_edges[0] == np.array([[2.0, 2.0, 2.0]])).all()
            assert (front_edges[1] == np.array([[6.0, 6.0, 6.0]])).all()

            front_edges = frame.parallel_front_edge_arrays_from_frame(image, rows=(0, 3))
            assert (front_edges[0] == np.array([[2.0, 2.0, 2.0],
                                                [3.0, 3.0, 3.0],
                                                [4.0, 4.0, 4.0]])).all()
            assert (front_edges[1] == np.array([[6.0, 6.0, 6.0],
                                                [7.0, 7.0, 7.0],
                                                [8.0, 8.0, 8.0]])).all()

    class TestParallelTrailsFromFrame:

        def test__pattern_bottom__extracts_1_trails_correctly(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 3, 0, 3)])

            image = np.array([[0.0, 0.0, 0.0],
                              [1.0, 1.0, 1.0],
                              [2.0, 2.0, 2.0],
                              [3.0, 3.0, 3.0],
                              # <- Trails form here onwards according to region and this frame_geometry
                              [4.0, 4.0, 4.0],  # <- Next trail.
                              [5.0, 5.0, 5.0],
                              [6.0, 6.0, 6.0],
                              [7.0, 7.0, 7.0],
                              [8.0, 8.0, 8.0],
                              [9.0, 9.0, 9.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(0, 1))

            assert (trails == np.array([[3.0, 3.0, 3.0]])).all()
            trails = frame.parallel_trails_arrays_from_frame(image, rows=(1, 2))

            assert (trails == np.array([[4.0, 4.0, 4.0]])).all()
            trails = frame.parallel_trails_arrays_from_frame(image, rows=(2, 3))
            assert (trails == np.array([[5.0, 5.0, 5.0]])).all()

        def test__pattern_bottom__extracts_multiple_trails_correctly(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 3, 0, 3)])

            image = np.array([[0.0, 0.0, 0.0],
                              [1.0, 1.0, 1.0],
                              [2.0, 2.0, 2.0],
                              [3.0, 3.0, 3.0],
                              # <- Trails form here onwards according to region and this frame_geometry
                              [4.0, 4.0, 4.0],  # <- Next trail.
                              [5.0, 5.0, 5.0],
                              [6.0, 6.0, 6.0],
                              [7.0, 7.0, 7.0],
                              [8.0, 8.0, 8.0],
                              [9.0, 9.0, 9.0]])
            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(0, 2))
            assert (trails == np.array([[3.0, 3.0, 3.0],
                                        [4.0, 4.0, 4.0]])).all()

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(1, 3))
            assert (trails == np.array([[4.0, 4.0, 4.0],
                                        [5.0, 5.0, 5.0]])).all()

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(1, 4))
            assert (trails == np.array([[4.0, 4.0, 4.0],
                                        [5.0, 5.0, 5.0],
                                        [6.0, 6.0, 6.0]])).all()

        def test__pattern_bottom__2_regions__extracts_rows_correctly(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 3, 0, 3), (4, 6, 0, 3)])

            image = np.array([[0.0, 0.0, 0.0],
                              [1.0, 1.0, 1.0],
                              [2.0, 2.0, 2.0],
                              [3.0, 3.0, 3.0],
                              # <- 1st Trails form here onwards according to region and this frame_geometry
                              [4.0, 4.0, 4.0],  # <- Next trail.
                              [5.0, 5.0, 5.0],
                              [6.0, 6.0, 6.0],
                              # <- 2nd Trails form here onwards according to region and this frame_geometry
                              [7.0, 7.0, 7.0],  # <- Next trail.
                              [8.0, 8.0, 8.0],
                              [9.0, 9.0, 9.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(0, 1))
            assert (trails[0] == np.array([[3.0, 3.0, 3.0]])).all()
            assert (trails[1] == np.array([[6.0, 6.0, 6.0]])).all()

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(0, 2))
            assert (trails[0] == np.array([[3.0, 3.0, 3.0],
                                           [4.0, 4.0, 4.0]])).all()
            assert (trails[1] == np.array([[6.0, 6.0, 6.0],
                                           [7.0, 7.0, 7.0]])).all()

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(1, 4))
            assert (trails[0] == np.array([[4.0, 4.0, 4.0],
                                           [5.0, 5.0, 5.0],
                                           [6.0, 6.0, 6.0]])).all()
            assert (trails[1] == np.array([[7.0, 7.0, 7.0],
                                           [8.0, 8.0, 8.0],
                                           [9.0, 9.0, 9.0]])).all()

        def test__pattern_top__does_all_the_above_correctly(self):

            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(3, 5, 0, 3), (9, 10, 0, 3)])

            image = np.array([[0.0, 0.0, 0.0],
                              [1.0, 1.0, 1.0],  # <- Next trail.
                              [2.0, 2.0, 2.0],
                              # <- 1st Trails form here onwards according to region and this frame_geometry
                              [3.0, 3.0, 3.0],
                              [4.0, 4.0, 4.0],
                              [5.0, 5.0, 5.0],
                              [6.0, 6.0, 6.0],
                              [7.0, 7.0, 7.0],  # <- Next trail.
                              [8.0, 8.0, 8.0],
                              # <- 2nd Trails form here onwards according to region and this frame_geometry
                              [9.0, 9.0, 9.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), ci_pattern=pattern)

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(0, 1))
            assert (trails[0] == np.array([[2.0, 2.0, 2.0]])).all()
            assert (trails[1] == np.array([[8.0, 8.0, 8.0]])).all()

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(0, 2))
            assert (trails[0] == np.array([[1.0, 1.0, 1.0],
                                           [2.0, 2.0, 2.0]])).all()
            assert (trails[1] == np.array([[7.0, 7.0, 7.0],
                                           [8.0, 8.0, 8.0]])).all()

            trails = frame.parallel_trails_arrays_from_frame(image, rows=(1, 3))
            assert (trails[0] == np.array([[0.0, 0.0, 0.0],
                                           [1.0, 1.0, 1.0]])).all()
            assert (trails[1] == np.array([[6.0, 6.0, 6.0],
                                           [7.0, 7.0, 7.0]])).all()

    class TestExtractSerialFrontEdges:

        def test__pattern_bottom___extracts_1_front_edge_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])

            #       /| Front Edge

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            front_edge = frame.serial_front_edge_arrays_from_frame(image, columns=(0, 1))

            assert (front_edge == np.array([[1.0],
                                            [1.0],
                                            [1.0]])).all()

            front_edge = frame.serial_front_edge_arrays_from_frame(image, columns=(1, 2))

            assert (front_edge == np.array([[2.0],
                                            [2.0],
                                            [2.0]])).all()

            front_edge = frame.serial_front_edge_arrays_from_frame(image, columns=(2, 3))

            assert (front_edge == np.array([[3.0],
                                            [3.0],
                                            [3.0]])).all()

        def test__pattern_bottom___extracts_multiple_front_edges_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 5)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])

            #                    /| Front Edge

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            front_edge = frame.serial_front_edge_arrays_from_frame(image, columns=(0, 2))

            assert (front_edge == np.array([[1.0, 2.0],
                                            [1.0, 2.0],
                                            [1.0, 2.0]])).all()

            front_edge = frame.serial_front_edge_arrays_from_frame(image, columns=(1, 4))

            assert (front_edge == np.array([[2.0, 3.0, 4.0],
                                            [2.0, 3.0, 4.0],
                                            [2.0, 3.0, 4.0]])).all()

        def test__pattern_bottom__2_regions__extracts_columns_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4), (0, 3, 5, 8)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])

            #                    /| FE 1        /\ FE 2

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            front_edges = frame.serial_front_edge_arrays_from_frame(image, columns=(0, 1))

            assert (front_edges[0] == np.array([[1.0],
                                                [1.0],
                                                [1.0]])).all()
            assert (front_edges[1] == np.array([[5.0],
                                                [5.0],
                                                [5.0]])).all()

            front_edges = frame.serial_front_edge_arrays_from_frame(image, columns=(1, 2))

            assert (front_edges[0] == np.array([[2.0],
                                                [2.0],
                                                [2.0]])).all()
            assert (front_edges[1] == np.array([[6.0],
                                                [6.0],
                                                [6.0]])).all()

            front_edges = frame.serial_front_edge_arrays_from_frame(image, columns=(2, 3))

            assert (front_edges[0] == np.array([[3.0],
                                                [3.0],
                                                [3.0]])).all()
            assert (front_edges[1] == np.array([[7.0],
                                                [7.0],
                                                [7.0]])).all()

            front_edges = frame.serial_front_edge_arrays_from_frame(image, columns=(0, 3))

            assert (front_edges[0] == np.array([[1.0, 2.0, 3.0],
                                                [1.0, 2.0, 3.0],
                                                [1.0, 2.0, 3.0]])).all()

            assert (front_edges[1] == np.array([[5.0, 6.0, 7.0],
                                                [5.0, 6.0, 7.0],
                                                [5.0, 6.0, 7.0]])).all()

        def test__pattern_right__does_all_the_above_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4), (0, 3, 5, 8)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])

            #                               /| FE 1            /\ FE 2

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

            front_edges = frame.serial_front_edge_arrays_from_frame(image, columns=(0, 1))

            assert (front_edges[0] == np.array([[3.0],
                                                [3.0],
                                                [3.0]])).all()
            assert (front_edges[1] == np.array([[7.0],
                                                [7.0],
                                                [7.0]])).all()

            front_edges = frame.serial_front_edge_arrays_from_frame(image, columns=(1, 2))

            assert (front_edges[0] == np.array([[2.0],
                                                [2.0],
                                                [2.0]])).all()
            assert (front_edges[1] == np.array([[6.0],
                                                [6.0],
                                                [6.0]])).all()

            front_edges = frame.serial_front_edge_arrays_from_frame(image, columns=(2, 3))

            assert (front_edges[0] == np.array([[1.0],
                                                [1.0],
                                                [1.0]])).all()
            assert (front_edges[1] == np.array([[5.0],
                                                [5.0],
                                                [5.0]])).all()

            front_edges = frame.serial_front_edge_arrays_from_frame(image, columns=(0, 3))

            assert (front_edges[0] == np.array([[1.0, 2.0, 3.0],
                                                [1.0, 2.0, 3.0],
                                                [1.0, 2.0, 3.0]])).all()

            assert (front_edges[1] == np.array([[5.0, 6.0, 7.0],
                                                [5.0, 6.0, 7.0],
                                                [5.0, 6.0, 7.0]])).all()

    class TestExtractSerialTrails:

        def test__pattern_bottom___extracts_1_trails_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])

            #                                    /| Trails Begin          

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            trails = frame.serial_trails_arrays_from_frame(image, columns=(0, 1))

            assert (trails == np.array([[4.0],
                                        [4.0],
                                        [4.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(1, 2))

            assert (trails == np.array([[5.0],
                                        [5.0],
                                        [5.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(2, 3))

            assert (trails == np.array([[6.0],
                                        [6.0],
                                        [6.0]])).all()

        def test__pattern_bottom___extracts_multiple_trails_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]])

            #                                   /| Trails Begin

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            trails = frame.serial_trails_arrays_from_frame(image, columns=(0, 2))

            assert (trails == np.array([[4.0, 5.0],
                                        [4.0, 5.0],
                                        [4.0, 5.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(1, 4))

            assert (trails == np.array([[5.0, 6.0, 7.0],
                                        [5.0, 6.0, 7.0],
                                        [5.0, 6.0, 7.0]])).all()

        def test__pattern_bottom__2_regions__extracts_columns_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4), (0, 3, 5, 8)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]])

            #                                   /| Trails1           /\ Trails2

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            trails = frame.serial_trails_arrays_from_frame(image, columns=(0, 1))

            assert (trails[0] == np.array([[4.0],
                                           [4.0],
                                           [4.0]])).all()
            assert (trails[1] == np.array([[8.0],
                                           [8.0],
                                           [8.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(1, 2))

            assert (trails[0] == np.array([[5.0],
                                           [5.0],
                                           [5.0]])).all()
            assert (trails[1] == np.array([[9.0],
                                           [9.0],
                                           [9.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(2, 3))

            assert (trails[0] == np.array([[6.0],
                                           [6.0],
                                           [6.0]])).all()
            assert (trails[1] == np.array([[10.0],
                                           [10.0],
                                           [10.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(0, 3))

            assert (trails[0] == np.array([[4.0, 5.0, 6.0],
                                           [4.0, 5.0, 6.0],
                                           [4.0, 5.0, 6.0]])).all()

            assert (trails[1] == np.array([[8.0, 9.0, 10.0],
                                           [8.0, 9.0, 10.0],
                                           [8.0, 9.0, 10.0]])).all()

        def test__pattern_right__does_all_the_above_correctly(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 3, 6), (0, 3, 8, 11)])

            image = np.array([[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0],
                              [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]])

            #               Trails1   /|                Trails2 /\

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

            trails = frame.serial_trails_arrays_from_frame(image, columns=(0, 1))

            assert (trails[0] == np.array([[2.0],
                                           [2.0],
                                           [2.0]])).all()
            assert (trails[1] == np.array([[7.0],
                                           [7.0],
                                           [7.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(1, 2))

            assert (trails[0] == np.array([[1.0],
                                           [1.0],
                                           [1.0]])).all()
            assert (trails[1] == np.array([[6.0],
                                           [6.0],
                                           [6.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(2, 3))

            assert (trails[0] == np.array([[0.0],
                                           [0.0],
                                           [0.0]])).all()
            assert (trails[1] == np.array([[5.0],
                                           [5.0],
                                           [5.0]])).all()

            trails = frame.serial_trails_arrays_from_frame(image, columns=(0, 3))

            assert (trails[0] == np.array([[0.0, 1.0, 2.0],
                                           [0.0, 1.0, 2.0],
                                           [0.0, 1.0, 2.0]])).all()

            assert (trails[1] == np.array([[5.0, 6.0, 7.0],
                                           [5.0, 6.0, 7.0],
                                           [5.0, 6.0, 7.0]])).all()

    class TestParallelSerialCalibrationSection:

        def test__extracts_everything_except_prescan(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 1, 0, 1)])

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [0.0, 1.0, 2.0, 3.0],
                              [0.0, 1.0, 2.0, 3.0],
                              [0.0, 1.0, 2.0, 3.0],
                              [0.0, 1.0, 2.0, 3.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            frame.frame_geometry.serial_prescan = ci_frame.Region(region=(0, 4, 0, 1))

            extracted_array = frame.parallel_serial_calibration_section(image)

            assert (extracted_array == np.array([[1.0, 2.0, 3.0],
                                                 [1.0, 2.0, 3.0],
                                                 [1.0, 2.0, 3.0],
                                                 [1.0, 2.0, 3.0],
                                                 [1.0, 2.0, 3.0]])).all()

        def test__extracts_everything_except_prescan_2(self):
            pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 1, 0, 1)])

            image = np.array([[0.0, 1.0, 2.0, 3.0],
                              [0.0, 1.0, 2.0, 3.0],
                              [0.0, 1.0, 2.0, 3.0],
                              [0.0, 1.0, 2.0, 3.0],
                              [0.0, 1.0, 2.0, 3.0]])

            frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

            frame.frame_geometry.serial_prescan = ci_frame.Region(region=(0, 4, 0, 2))

            extracted_array = frame.parallel_serial_calibration_section(image)

            assert (extracted_array == np.array([[2.0, 3.0],
                                                 [2.0, 3.0],
                                                 [2.0, 3.0],
                                                 [2.0, 3.0],
                                                 [2.0, 3.0]])).all()


class TestQuadGeometryEuclid_bottom_left(object):
    class TestParallelFrontEdgeRegion:

        def test__extract_one_row__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region(region=(0, 3, 0, 3))  # Front edge is row 0, so for 1 row we extract 0 -> 1

            ci_front_edge = ci_geometry.parallel_front_edge_region(ci_region=ci_region, rows=(0, 1))

            assert ci_front_edge == (0, 1, 0, 3)

        def test__extract_two_rows__first_and_second__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region(region=(0, 3, 0, 3))  # Front edge is row 0, so for 2 rows we extract 0 -> 2

            ci_front_edge = ci_geometry.parallel_front_edge_region(ci_region=ci_region, rows=(0, 2))

            assert ci_front_edge == (0, 2, 0, 3)

        def test__extract_two_rows__second_and_third__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is row 0, so for these 2 rows we extract 1 ->2

            ci_front_edge = ci_geometry.parallel_front_edge_region(ci_region=ci_region, rows=(1, 3))

            assert ci_front_edge == (1, 3, 0, 3)

    class TestParallelTrailsRegion:

        def test__extract_one_row__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region(region=(0, 3, 0, 3))  # The trails are row 3 and above, so extract 3 -> 4

            ci_trails = ci_geometry.parallel_trails_region(ci_region=ci_region, rows=(0, 1))

            assert ci_trails == (3, 4, 0, 3)

        def test__extract_two_rows__first_and_second__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region((0, 3, 0, 3))  # The trails are row 3 and above, so extract 3 -> 5

            ci_trails = ci_geometry.parallel_trails_region(ci_region=ci_region, rows=(0, 2))

            assert ci_trails == (3, 5, 0, 3)

        def test__extract_two_rows__second_and_third__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region((0, 3, 0, 3))  # The trails are row 3 and above, so extract 4 -> 6

            ci_trails = ci_geometry.parallel_trails_region(ci_region=ci_region, rows=(1, 3))

            assert ci_trails == (4, 6, 0, 3)

    class TestParallelSideNearestReadOut:

        def test__columns_0_to_1__region_is_left_hand_side(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            ci_region = ci_frame.Region(region=(1, 3, 0, 5))

            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region, image_shape=(5, 5),
                                                                                   columns=(0, 1))

            assert ci_parallel_region == (0, 5, 0, 1)

        def test__columns_1_to_3__region_is_left_hand_side(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            ci_region = ci_frame.Region(region=(1, 3, 0, 5))

            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(4, 4), columns=(1, 3))

            assert ci_parallel_region == (0, 4, 1, 3)

        def test__columns_1_to_3__different_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            ci_region = ci_frame.Region(region=(1, 3, 2, 5))

            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(4, 4), columns=(1, 3))

            assert ci_parallel_region == (0, 4, 3, 5)

        def test__columns_0_to_1__asymetric_image(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            ci_region = ci_frame.Region(region=(1, 3, 0, 5))

            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(2, 5), columns=(0, 1))

            assert ci_parallel_region == (0, 2, 0, 1)

    class TestSerialFrontEdgeRegion:

        def test__extract_one_column__takes_coordinates_from_left_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is column 0, so for 1 column we extract 0 -> 1

            ci_front_edge = ci_geometry.serial_front_edge_region(ci_region=ci_region, columns=(0, 1))

            assert ci_front_edge == (0, 3, 0, 1)

        def test__extract_two_columns__first_and_second__takes_coordinates_from_left_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is column 0, so for 2 columns we extract 0 -> 2

            ci_front_edge = ci_geometry.serial_front_edge_region(ci_region=ci_region, columns=(0, 2))

            assert ci_front_edge == (0, 3, 0, 2)

        def test__extract_two_columns__second_and_third__takes_coordinates_from_left_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is column 0, so for these 2 columns we extract 1 ->2

            ci_front_edge = ci_geometry.serial_front_edge_region(ci_region=ci_region, columns=(1, 3))

            assert ci_front_edge == (0, 3, 1, 3)

    class TestSerialTrailsRegion:

        def test__extract_one_row__takes_coordinates_after_right_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region(region=(0, 3, 0, 3))  # The trails are column 3 and above, so extract 3 -> 4

            ci_trails = ci_geometry.serial_trails_region(ci_region=ci_region, columns=(0, 1))

            assert ci_trails == (0, 3, 3, 4)

        def test__extract_two_columns__first_and_second__takes_coordinates_after_right_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region((0, 3, 0, 3))  # The trails are column 3 and above, so extract 3 -> 5

            ci_trails = ci_geometry.serial_trails_region(ci_region=ci_region, columns=(0, 2))

            assert ci_trails == (0, 3, 3, 5)

        def test__extract_two_columns__second_and_third__takes_coordinates_after_right_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()

            ci_region = ci_frame.Region((0, 3, 0, 3))  # The trails are column 3 and above, so extract 4 -> 6

            ci_trails = ci_geometry.serial_trails_region(ci_region=ci_region, columns=(1, 3))

            assert ci_trails == (0, 3, 4, 6)

    class TestSerialChargeInjectionAndTrails:

        def test__column_0_of_front_edge__region_is_left_hand_side__no_overscan_beyond_ci_region(self):

            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            ci_region = ci_frame.Region(region=(1, 3, 0, 5))

            ci_serial_region = ci_geometry.serial_prescan_ci_region_and_trails(ci_region=ci_region, image_shape=(5, 5))

            assert ci_serial_region == (1, 3, 0, 5)

        def test__column_0_of_front_edge__region_is_left_hand_side__overscan_beyond_ci_region(self):

            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            ci_region = ci_frame.Region(region=(1, 3, 0, 5))

            ci_serial_region = ci_geometry.serial_prescan_ci_region_and_trails(ci_region=ci_region, image_shape=(5, 25))

            assert ci_serial_region == (1, 3, 0, 25)


class TestQuadGeometryEuclid_bottom_right(object):
    class TestParallelFrontEdgeOfRegion:

        def test__extract_two_rows__second_and_third__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is row 0, so for these 2 rows we extract 1 ->2

            ci_front_edge = ci_geometry.parallel_front_edge_region(ci_region=ci_region, rows=(1, 3))

            assert ci_front_edge == (1, 3, 0, 3)

    class TestParallelTrailsRegion:

        def test__extract_two_rows__second_and_third__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region((0, 3, 0, 3))  # The trails are row 3 and above, so extract 4 -> 6

            ci_trails = ci_geometry.parallel_trails_region(ci_region=ci_region, rows=(1, 3))

            assert ci_trails == (4, 6, 0, 3)

    class TestParallelSideNearestReadOut:

        def test__columns_0_to_1__region_is_right_hand_side(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            ci_region = ci_frame.Region(region=(1, 3, 0, 5))

            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(5, 5), columns=(0, 1))

            assert ci_parallel_region == (0, 5, 4, 5)

        def test__columns_1_to_3__region_is_right_hand_side(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            ci_region = ci_frame.Region(region=(1, 3, 0, 4))

            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(4, 4), columns=(1, 3))

            assert ci_parallel_region == (0, 4, 1, 3)

        def test__columns_1_to_3__different_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            ci_region = ci_frame.Region(region=(1, 3, 2, 4))

            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(4, 4), columns=(1, 3))

            assert ci_parallel_region == (0, 4, 1, 3)

        def test__columns_0_to_1__asymetric_image(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            ci_region = ci_frame.Region(region=(0, 1, 0, 5))
            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(2, 5), columns=(0, 1))

            assert ci_parallel_region == (0, 2, 4, 5)

    class TestSerialFrontEdgeRegion:

        def test__extract_one_column__takes_coordinates_from_right_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is column 0, so for 1 column we extract 0 -> 1

            ci_front_edge = ci_geometry.serial_front_edge_region(ci_region=ci_region, columns=(0, 1))

            assert ci_front_edge == (0, 3, 2, 3)

        def test__extract_two_columns__first_and_second__takes_coordinates_from_right_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is column 0, so for 2 columns we extract 0 -> 2

            ci_front_edge = ci_geometry.serial_front_edge_region(ci_region=ci_region, columns=(0, 2))

            assert ci_front_edge == (0, 3, 1, 3)

        def test__extract_two_columns__second_and_third__takes_coordinates_from_right_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is column 0, so for these 2 columns we extract 1 ->2

            ci_front_edge = ci_geometry.serial_front_edge_region(ci_region=ci_region, columns=(1, 3))

            assert ci_front_edge == (0, 3, 0, 2)

    class TestSerialTrailsRegion:

        def test__extract_one_row__takes_coordinates_after_left_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region(region=(0, 3, 3, 6))  # The trails are column 3 and above, so extract 3 -> 4

            ci_trails = ci_geometry.serial_trails_region(ci_region=ci_region, columns=(0, 1))

            assert ci_trails == (0, 3, 2, 3)

        def test__extract_two_columns__first_and_second__takes_coordinates_after_left_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region((0, 3, 3, 6))  # The trails are column 3 and above, so extract 3 -> 5

            ci_trails = ci_geometry.serial_trails_region(ci_region=ci_region, columns=(0, 2))

            assert ci_trails == (0, 3, 1, 3)

        def test__extract_two_columns__second_and_third__takes_coordinates_after_left_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region((0, 3, 3, 6))  # The trails are column 3 and above, so extract 4 -> 6

            ci_trails = ci_geometry.serial_trails_region(ci_region=ci_region, columns=(1, 3))

            assert ci_trails == (0, 3, 0, 2)

    class TestSerialChargeInjectionAndTrails:

        def test__front_edge__region_is_left_hand_side__no_overscan_beyond_ci_region(self):

            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            ci_region = ci_frame.Region(region=(1, 3, 0, 5))

            ci_serial_region = ci_geometry.serial_prescan_ci_region_and_trails(ci_region=ci_region, image_shape=(5, 5))

            assert ci_serial_region == (1, 3, 0, 5)

        def test__front_edge__region_is_left_hand_side__overscan_beyond_ci_region(self):

            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            ci_region = ci_frame.Region(region=(1, 3, 20, 25))

            ci_serial_region = ci_geometry.serial_prescan_ci_region_and_trails(ci_region=ci_region, image_shape=(5, 25))

            assert ci_serial_region == (1, 3, 0, 25)

        def test__ci_region_has_overscan_and_prescan_either_side__prescan_included(self):

            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()
            ci_region = ci_frame.Region(region=(1, 3, 10, 20))

            ci_serial_region = ci_geometry.serial_prescan_ci_region_and_trails(ci_region=ci_region, image_shape=(5, 25))

            assert ci_serial_region == (1, 3, 0, 25)


class TestQuadGeometryEuclid_top_left(object):
    class TestParallelFrontEdgeOfRegion:

        def test__extract_one_row__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()

            ci_region = ci_frame.Region(
                (0, 3, 0, 3))  # The front edge is closest to 3, so for 1 edge we extract row 3-> 4

            ci_front_edge = ci_geometry.parallel_front_edge_region(ci_region=ci_region, rows=(0, 1))

            assert ci_front_edge == (2, 3, 0, 3)

        def test__extract_two_rows__first_and_second__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()

            ci_region = ci_frame.Region(
                (0, 3, 0, 3))  # The front edge is closest to 3, so for these 2 rows we extract 2 & 3

            ci_front_edge = ci_geometry.parallel_front_edge_region(ci_region=ci_region, rows=(0, 2))

            assert ci_front_edge == (1, 3, 0, 3)

        def test__extract_two_rows__second_and_third__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()

            ci_region = ci_frame.Region(
                (0, 3, 0, 3))  # The front edge is closest to 3, so for these 2 rows we extract 1 & 2

            ci_front_edge = ci_geometry.parallel_front_edge_region(ci_region=ci_region, rows=(1, 3))

            assert ci_front_edge == (0, 2, 0, 3)

    class TestParallelTrailsRegion:

        def test__extract_one_row__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()

            ci_region = ci_frame.Region(
                (3, 5, 0, 3))  # The trails are the rows after row 3, so for 1 edge we should extract just row 2

            ci_trails = ci_geometry.parallel_trails_region(ci_region=ci_region, rows=(0, 1))

            assert ci_trails == (2, 3, 0, 3)

        def test__extract_two_rows__first_and_second__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()

            ci_region = ci_frame.Region(
                (3, 5, 0, 3))  # The trails are the row after row 3, so for these 2 edges we extract rows 1->3

            ci_trails = ci_geometry.parallel_trails_region(ci_region=ci_region, rows=(0, 2))

            assert ci_trails == (1, 3, 0, 3)

        def test__extract_two_rows__second_and_third__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()

            ci_region = ci_frame.Region(
                (3, 5, 0, 3))  # The trails are the row after row 3, so for these 2 edges we extract rows 0 & 2

            ci_trails = ci_geometry.parallel_trails_region(ci_region=ci_region, rows=(1, 3))

            assert ci_trails == (0, 2, 0, 3)

    class TestParallelSideNearestReadOut:

        def test__columns_0_to_1__asymetric_image(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_left()
            ci_region = ci_frame.Region(region=(1, 3, 0, 5))

            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(2, 5), columns=(0, 1))

            assert ci_parallel_region == (0, 2, 0, 1)

    class TestSerialFrontEdgeRegion:

        def test__extract_two_columns__second_and_third__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is column 0, so for these 2 columns we extract 1 ->2

            ci_front_edge = ci_geometry.serial_front_edge_region(ci_region=ci_region, columns=(1, 3))

            assert ci_front_edge == (0, 3, 1, 3)

    class TestSerialTrailsRegion:

        def test__extract_two_columns__second_and_third__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()

            ci_region = ci_frame.Region((0, 3, 0, 3))  # The trails are column 3 and above, so extract 4 -> 6

            ci_trails = ci_geometry.serial_trails_region(ci_region=ci_region, columns=(1, 3))

            assert ci_trails == (0, 3, 4, 6)

    class TestSerialChargeInjectionAndTrails:

        def test__different_ci_region(self):

            ci_geometry = ci_frame.QuadGeometryEuclid.top_left()
            ci_region = ci_frame.Region(region=(3, 5, 5, 30))

            ci_serial_region = ci_geometry.serial_prescan_ci_region_and_trails(ci_region=ci_region, image_shape=(8, 55))

            assert ci_serial_region == (3, 5, 0, 55)


class TestQuadGeometryEuclid_top_right(object):
    class TestParallelFrontEdgeOfRegion:

        def test__extract_two_rows__second_and_third__takes_coordinates_from_top_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_right()

            ci_region = ci_frame.Region(
                (0, 3, 0, 3))  # The front edge is closest to 3, so for these 2 rows we extract 1 & 2

            ci_front_edge = ci_geometry.parallel_front_edge_region(ci_region=ci_region, rows=(1, 3))

            assert ci_front_edge == (0, 2, 0, 3)

    class TestParallelTrailsRegion:

        def test__extract_two_rows__second_and_third__takes_coordinates_after_bottom_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_right()

            ci_region = ci_frame.Region(
                (3, 5, 0, 3))  # The trails are the row after row 3, so for these 2 edges we extract rows 0 & 2

            ci_trails = ci_geometry.parallel_trails_region(ci_region=ci_region, rows=(1, 3))

            assert ci_trails == (0, 2, 0, 3)

    class TestParallelSideNearestReadOut:

        def test__columns_0_to_1__asymetric_image(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.top_right()
            ci_region = ci_frame.Region(region=(0, 1, 0, 5))
            ci_parallel_region = ci_geometry.parallel_side_nearest_read_out_region(ci_region=ci_region,
                                                                                   image_shape=(2, 5), columns=(0, 1))

            assert ci_parallel_region == (0, 2, 4, 5)

    class TestSerialFrontEdgeRegion:

        def test__extract_two_columns__second_and_third__takes_coordinates_from_right_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region(
                region=(0, 3, 0, 3))  # Front edge is column 0, so for these 2 columns we extract 1 ->2

            ci_front_edge = ci_geometry.serial_front_edge_region(ci_region=ci_region, columns=(1, 3))

            assert ci_front_edge == (0, 3, 0, 2)

    class TestSerialTrailsRegion:

        def test__extract_two_columns__second_and_third__takes_coordinates_after_left_of_region(self):
            ci_geometry = ci_frame.QuadGeometryEuclid.bottom_right()

            ci_region = ci_frame.Region((0, 3, 3, 6))  # The trails are column 3 and above, so extract 4 -> 6

            ci_trails = ci_geometry.serial_trails_region(ci_region=ci_region, columns=(1, 3))

            assert ci_trails == (0, 3, 0, 2)

    class TestSerialChargeInjectionAndTrails:

        def test__different_ci_region(self):

            ci_geometry = ci_frame.QuadGeometryEuclid.top_right()
            ci_region = ci_frame.Region(region=(3, 5, 5, 30))

            ci_serial_region = ci_geometry.serial_prescan_ci_region_and_trails(ci_region=ci_region, image_shape=(8, 55))

            assert ci_serial_region == (3, 5, 0, 55)



from autocti.pipeline import tagging

class TestPhaseTag:

    def test__mixture_of_values(self):

        phase_tag = tagging.phase_tag_from_phase_settings(columns=1, rows=(0,1),
                                                          parallel_front_edge_mask_rows=None,
                                                          parallel_trails_mask_rows=None,
                                                          serial_front_edge_mask_columns=None,
                                                          serial_trails_mask_columns=None,
                                                          parallel_total_density_range=None,
                                                          serial_total_density_range=None,
                                                          cosmic_ray_parallel_buffer=None,
                                                          cosmic_ray_serial_buffer=None,
                                                          cosmic_ray_diagonal_buffer=None)


        assert phase_tag == '_col_1_rows_(0,1)'

        phase_tag = tagging.phase_tag_from_phase_settings(columns=1, rows=(0,1),
                                                          parallel_front_edge_mask_rows=(0,1),
                                                          parallel_trails_mask_rows=None,
                                                          serial_front_edge_mask_columns=None,
                                                          serial_trails_mask_columns=(5,10),
                                                          parallel_total_density_range=None,
                                                          serial_total_density_range=None,
                                                          cosmic_ray_parallel_buffer=None,
                                                          cosmic_ray_serial_buffer=None,
                                                          cosmic_ray_diagonal_buffer=None)


        assert phase_tag == '_col_1_rows_(0,1)_par_front_mask_rows_(0,1)_ser_trails_mask_col_(5,10)'

        phase_tag = tagging.phase_tag_from_phase_settings(columns=1, rows=(0,1),
                                                          parallel_front_edge_mask_rows=None,
                                                          parallel_trails_mask_rows=None,
                                                          serial_front_edge_mask_columns=None,
                                                          serial_trails_mask_columns=None,
                                                          parallel_total_density_range=None,
                                                          serial_total_density_range=None,
                                                          cosmic_ray_parallel_buffer=1,
                                                          cosmic_ray_serial_buffer=2,
                                                          cosmic_ray_diagonal_buffer=3)


        assert phase_tag == '_col_1_rows_(0,1)_cr_p1s2d3'

        phase_tag = tagging.phase_tag_from_phase_settings(columns=2, rows=(1,2),
                                                          parallel_front_edge_mask_rows=None,
                                                          parallel_trails_mask_rows=None,
                                                          serial_front_edge_mask_columns=None,
                                                          serial_trails_mask_columns=None,
                                                          parallel_total_density_range=None,
                                                          serial_total_density_range=None,
                                                          cosmic_ray_parallel_buffer=4,
                                                          cosmic_ray_serial_buffer=5,
                                                          cosmic_ray_diagonal_buffer=6)

        assert phase_tag == '_col_2_rows_(1,2)_cr_p4s5d6'

        phase_tag = tagging.phase_tag_from_phase_settings(columns=2, rows=(1,2),
                                                          parallel_front_edge_mask_rows=None,
                                                          parallel_trails_mask_rows=None,
                                                          serial_front_edge_mask_columns=None,
                                                          serial_trails_mask_columns=None,
                                                          parallel_total_density_range=(0, 1),
                                                          serial_total_density_range=(2,3),
                                                          cosmic_ray_parallel_buffer=4,
                                                          cosmic_ray_serial_buffer=5,
                                                          cosmic_ray_diagonal_buffer=6)

        assert phase_tag == '_col_2_rows_(1,2)_par_range_(0,1)_ser_range_(2,3)_cr_p4s5d6'

        phase_tag = tagging.phase_tag_from_phase_settings(columns=1, rows=(0,1),
                                                          parallel_front_edge_mask_rows=None,
                                                          parallel_trails_mask_rows=(1,2),
                                                          serial_front_edge_mask_columns=(2,4),
                                                          serial_trails_mask_columns=None,
                                                          parallel_total_density_range=None,
                                                          serial_total_density_range=None,
                                                          cosmic_ray_parallel_buffer=4,
                                                          cosmic_ray_serial_buffer=5,
                                                          cosmic_ray_diagonal_buffer=6)

        assert phase_tag == '_col_1_rows_(0,1)_par_trails_mask_rows_(1,2)_ser_front_mask_col_(2,4)_cr_p4s5d6'


class TestTaggers:

    def test__columns_tagger(self):

        tag = tagging.columns_tag_from_columns(columns=None)
        assert tag == ''
        tag = tagging.columns_tag_from_columns(columns=10)
        assert tag == '_col_10'
        tag = tagging.columns_tag_from_columns(columns=60)
        assert tag == '_col_60'
        
    def test__rows_tagger(self):

        tag = tagging.rows_tag_from_rows(rows=None)
        assert tag == ''
        tag = tagging.rows_tag_from_rows(rows=(0, 5))
        assert tag == '_rows_(0,5)'
        tag = tagging.rows_tag_from_rows(rows=(10, 20))
        assert tag == '_rows_(10,20)'
        
    def test__parallel_front_edge_mask_rows_tagger(self):

        tag = tagging.parallel_front_edge_mask_rows_tag_from_parallel_front_edge_mask_rows(
            parallel_front_edge_mask_rows=None)
        assert tag == ''
        tag = tagging.parallel_front_edge_mask_rows_tag_from_parallel_front_edge_mask_rows(
            parallel_front_edge_mask_rows=(0, 5))
        assert tag == '_par_front_mask_rows_(0,5)'
        tag = tagging.parallel_front_edge_mask_rows_tag_from_parallel_front_edge_mask_rows(
            parallel_front_edge_mask_rows=(10, 20))
        assert tag == '_par_front_mask_rows_(10,20)'

    def test__parallel_trails_mask_rows_tagger(self):
        
        tag = tagging.parallel_trails_mask_rows_tag_from_parallel_trails_mask_rows(
            parallel_trails_mask_rows=None)
        assert tag == ''

        tag = tagging.parallel_trails_mask_rows_tag_from_parallel_trails_mask_rows(
            parallel_trails_mask_rows=(0, 5))
        assert tag == '_par_trails_mask_rows_(0,5)'
        tag = tagging.parallel_trails_mask_rows_tag_from_parallel_trails_mask_rows(
            parallel_trails_mask_rows=(10, 20))
        assert tag == '_par_trails_mask_rows_(10,20)'

    def test__serial_front_edge_mask_columns_tagger(self):

        tag = tagging.serial_front_edge_mask_columns_tag_from_serial_front_edge_mask_columns(
            serial_front_edge_mask_columns=None)
        assert tag == ''

        tag = tagging.serial_front_edge_mask_columns_tag_from_serial_front_edge_mask_columns(
            serial_front_edge_mask_columns=(0, 5))
        assert tag == '_ser_front_mask_col_(0,5)'

        tag = tagging.serial_front_edge_mask_columns_tag_from_serial_front_edge_mask_columns(
            serial_front_edge_mask_columns=(10, 20))
        assert tag == '_ser_front_mask_col_(10,20)'

    def test__serial_trails_mask_columns_tagger(self):

        tag = tagging.serial_trails_mask_columns_tag_from_serial_trails_mask_columns(
            serial_trails_mask_columns=None)
        assert tag == ''

        tag = tagging.serial_trails_mask_columns_tag_from_serial_trails_mask_columns(
            serial_trails_mask_columns=(0, 5))
        assert tag == '_ser_trails_mask_col_(0,5)'

        tag = tagging.serial_trails_mask_columns_tag_from_serial_trails_mask_columns(
            serial_trails_mask_columns=(10, 20))
        assert tag == '_ser_trails_mask_col_(10,20)'

    def test__parallel_total_density_range_tagger(self):

        tag = tagging.parallel_total_density_range_tag_from_parallel_total_density_range(
            parallel_total_density_range=None)
        assert tag == ''
        tag = tagging.parallel_total_density_range_tag_from_parallel_total_density_range(
            parallel_total_density_range=(0, 5))
        assert tag == '_par_range_(0,5)'
        tag = tagging.parallel_total_density_range_tag_from_parallel_total_density_range(
            parallel_total_density_range=(10, 20))
        assert tag == '_par_range_(10,20)'
        
    def test__serial_total_density_range_tagger(self):

        tag = tagging.serial_total_density_range_tag_from_serial_total_density_range(
            serial_total_density_range=None)
        assert tag == ''
        tag = tagging.serial_total_density_range_tag_from_serial_total_density_range(
            serial_total_density_range=(0, 5))
        assert tag == '_ser_range_(0,5)'
        tag = tagging.serial_total_density_range_tag_from_serial_total_density_range(
            serial_total_density_range=(10, 20))
        assert tag == '_ser_range_(10,20)'

    def test__cosmic_ray_buffer_tagger(self):

        tag = tagging.cosmic_ray_buffer_tag_from_cosmic_ray_buffers(cosmic_ray_parallel_buffer=None,
                                                                    cosmic_ray_serial_buffer=None,
                                                                    cosmic_ray_diagonal_buffer=None)
        assert tag == ''

        tag = tagging.cosmic_ray_buffer_tag_from_cosmic_ray_buffers(cosmic_ray_parallel_buffer=1,
                                                                    cosmic_ray_serial_buffer=None,
                                                                    cosmic_ray_diagonal_buffer=3)
        assert tag == '_cr_p1d3'

        tag = tagging.cosmic_ray_buffer_tag_from_cosmic_ray_buffers(cosmic_ray_parallel_buffer=10,
                                                                    cosmic_ray_serial_buffer=20,
                                                                    cosmic_ray_diagonal_buffer=None)
        assert tag == '_cr_p10s20'

        tag = tagging.cosmic_ray_buffer_tag_from_cosmic_ray_buffers(cosmic_ray_parallel_buffer=1,
                                                                    cosmic_ray_serial_buffer=2,
                                                                    cosmic_ray_diagonal_buffer=3)
        assert tag == '_cr_p1s2d3'

        tag = tagging.cosmic_ray_buffer_tag_from_cosmic_ray_buffers(cosmic_ray_parallel_buffer=10,
                                                                    cosmic_ray_serial_buffer=5,
                                                                    cosmic_ray_diagonal_buffer=1)
        assert tag == '_cr_p10s5d1'


from autofit.core import model_mapper as mm
from autocti.model import arctic_params as ap
import workspace_jam.scripts.requirements as req

import matplotlib.pyplot as plt

model_mapper = mm.ModelMapper(serial=ap.SerialThreeSpecies)
model_mapper.serial.well_fill_alpha = 0.0
model_mapper.serial.well_fill_gamma = 0.0

input_model = [0.01, 0.03, 0.9, 0.8, 3.5, 20.0, 1.0e-4, 0.58]
requirements = [1.0, 0.0000631, 0.0084, 0.0039, 0.000303, 0.0193, 0.0300, 0.0004]
sigma_limit = 2.0

most_probable_lr, lower_limits_lr, upper_limits_lr, lower_error_lr, upper_error_lr, error_pecision_lr = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Serial_Tight_lr/')

most_probable_mr, lower_limits_mr, upper_limits_mr, lower_error_mr, upper_error_mr, error_pecision_mr = \
    req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model, sigma_limit=sigma_limit,
                                name='Results/Serial_Tight_mr/')

most_probable_hr, lower_limits_hr, upper_limits_hr, lower_error_hr, upper_error_hr, error_pecision_hr = \
  req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model,
                                                sigma_limit=sigma_limit, name='Results/Serial_Tight_hr/')

most_probable_x2, lower_limits_x2, upper_limits_x2, lower_error_x2, upper_error_x2,  error_pecision_x2 = \
req.setup_multinest_results(model_mapper=model_mapper, input_model=input_model,
                                               sigma_limit=sigma_limit, name='Results/Serial_Tight_x2/')

ylabels = [r'$\rho_{1}$', r'$\rho_{2}$', r'$\rho_{3}$', r'$\tau_{1}$', r'$\tau_{2}$', r'$\tau_{3}$', 'd', r'$\beta$',]

plt.figure(figsize=(20, 15))
plt.suptitle('Accuracy of Serial CTI Model', fontsize=20)

for i in range(8):
    plt.subplot(2, 4, i+1)
    plt.plot([496, 993, 1986, 3972], [input_model[i], input_model[i], input_model[i], input_model[i]], linestyle='--')
    print(i, most_probable_hr[i], (lower_error_hr[i], upper_error_hr[i]))
    plt.errorbar(x=[496, 993, 1986, 3972], y=[most_probable_lr[i], most_probable_mr[i],
                                              most_probable_hr[i], most_probable_x2[i]],
                 yerr=[[lower_error_lr[i], lower_error_mr[i], lower_error_hr[i], lower_error_x2[i]],
                       [upper_error_lr[i], upper_error_mr[i], upper_error_hr[i], upper_error_x2[i]]])
    plt.xlabel('Number of columns', fontsize=12)
    plt.ylabel(ylabels[i], fontsize=12)

plt.show()

plt.figure(figsize=(20, 15))
plt.suptitle('Precision of Serial CTI Model', fontsize=20)

for i in range(8):
    plt.subplot(2, 4, i+1)
    plt.plot([496, 993, 1986, 3972], [error_pecision_lr[i], error_pecision_mr[i],
                                      error_pecision_hr[i], error_pecision_x2[i]])
    plt.xlabel('Number of columns', fontsize=12)
    plt.ylabel(ylabels[i], fontsize=12)

plt.show()



#
# Copyright (C) 2012-2020 Euclid Science Ground Segment
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation; either version 3.0 of the License, or (at your option)
# any later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""
File: tests/python/CTICIData_test.py

Created on: 02/14/18
Author: user
"""

import numpy as np

from autocti.charge_injection import ci_frame
from autocti.charge_injection import ci_pattern
from autocti.charge_injection import ci_mask


class MockPattern(object):

    def __init__(self):
        pass


class TestMaskedParallelFrontEdge:

    def test__pattern_bottom___mask_only_contains_front_edge(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 3)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_parallel_front_edge_from_ci_frame(shape=(10, 3), ci_frame=frame, rows=(0, 2))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False,  False,  False],
                                  [True,  True,  True],  # <- Front edge according to region and this frame_geometry
                                  [True,  True,  True],  # <- Next front edge row.
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False]])).all()


    def test__same_as_above_but_uses_invert(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 3)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_parallel_front_edge_from_ci_frame(shape=(10, 3), ci_frame=frame, rows=(0, 2), 
                                                                       invert=True)

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[True,  True,  True],
                                  [False, False, False],  # <- Front edge according to region and this frame_geometry
                                  [False, False, False],  # <- Next front edge row.
                                  [True,  True,  True],
                                  [True,  True,  True],
                                  [True,  True,  True],
                                  [True,  True,  True],
                                  [True,  True,  True],
                                  [True,  True,  True],
                                  [True,  True,  True]])).all()
        
    def test__pattern_bottom__2_regions__extracts_rows_correctly(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 1), (1, 4, 2, 3)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_parallel_front_edge_from_ci_frame(shape=(10, 3), ci_frame=frame, rows=(0, 2))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False,  False,  False],
                                  [True, False, True],  # <- Front edge according to region and this frame_geometry
                                  [True, False, True],  # <- Next front edge row.
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False]])).all()

    def test__pattern_top__mask_only_contains_front_edge(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 3)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_parallel_front_edge_from_ci_frame(shape=(10, 3), ci_frame=frame, rows=(0, 2))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False,  False,  False],
                                  [False,   False,  False],
                                  [True,  True,  True],  # <- Next front edge row.
                                  [True,  True,  True], # <- Front edge according to region and this frame_geometry
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False],
                                  [False,  False,  False]])).all()


class TestMaskedParallelTrails:

    def test__pattern_bottom___mask_only_contains_trails(self):
        
        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 3)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_parallel_trails_from_ci_frame(shape=(10, 3), ci_frame=frame, rows=(0, 4))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False, False, False],
                                  [False, False, False],  
                                  [False, False, False],  
                                  [False, False, False],
                                  [True,  True,  True], # <- Frist Trail according to region and this frame_geometry
                                  [True,  True,  True], # <- Next trail row.
                                  [True,  True,  True],
                                  [True,  True,  True],
                                  [False, False, False],
                                  [False, False, False]])).all()

    def test__same_as_above_but_uses_invert(self):
        
        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 3)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_parallel_trails_from_ci_frame(shape=(10, 3), ci_frame=frame, rows=(0, 4), 
                                                                   invert=True)

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[True, True, True],
                                  [True, True, True],
                                  [True, True, True],
                                  [True, True, True],
                                  [False, False, False],  # <- Frist Trail according to region and this frame_geometry
                                  [False, False, False],  # <- Next trail row.
                                  [False, False, False],
                                  [False, False, False],
                                  [True, True, True],
                                  [True, True, True]])).all()

    def test__pattern_bottom__2_regions__extracts_rows_correctly(self):
        
        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 1), (1, 4, 2, 3)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_parallel_trails_from_ci_frame(shape=(10, 3), ci_frame=frame, rows=(0, 4))

        assert type(mask) == ci_mask.CIMask
        
        assert (mask == np.array([[False, False, False],
                                  [False, False, False],
                                  [False, False, False],
                                  [False, False, False],
                                  [True,  False,  True], # <- Frist Trail according to region and this frame_geometry
                                  [True,  False,  True], # <- Next trail row.
                                  [True,  False,  True],
                                  [True,  False,  True],
                                  [False, False, False],
                                  [False, False, False]])).all()

    def test__pattern_top__mask_only_contains_trails(self):
        
        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(1, 4, 0, 3)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.top_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_parallel_trails_from_ci_frame(shape=(10, 3), ci_frame=frame, rows=(0, 1))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[True,  True,  True], # <- Frist Trail according to region and this frame_geometry
                                  [False, False, False],
                                  [False, False, False], 
                                  [False, False, False], 
                                  [False, False, False], 
                                  [False, False, False],
                                  [False, False, False],
                                  [False, False, False],
                                  [False, False, False],
                                  [False, False, False]])).all()


class TestMaskedSerialFrontEdge:

    def test__pattern_left___mask_only_contains_front_edge(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_serial_front_edge_from_ci_frame(shape=(3, 10), ci_frame=frame, columns=(0, 2))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False, True, True, False, False, False, False, False, False, False],
                                  [False, True, True, False, False, False, False, False, False, False],
                                  [False, True, True, False, False, False, False, False, False, False]])).all()

    def test__same_as_above_but_uses_invert(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_serial_front_edge_from_ci_frame(shape=(3, 10), ci_frame=frame, columns=(0, 2),
                                                                     invert=True)

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[True, False, False, True, True, True, True, True, True, True],
                                  [True, False, False, True, True, True, True, True, True, True],
                                  [True, False, False, True, True, True, True, True, True, True]])).all()

    def test__pattern_left__2_regions__extracts_columns_correctly(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 1, 1, 4), (2, 3, 1, 4)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_serial_front_edge_from_ci_frame(shape=(3, 10), ci_frame=frame, columns=(0, 3))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False, True, True, True, False, False, False, False, False, False],
                                  [False,  False,  False,  False, False, False, False, False, False, False],
                                  [False, True, True, True, False, False, False, False, False, False]])).all()

    def test__pattern_right__mask_only_contains_front_edge(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_serial_front_edge_from_ci_frame(shape=(3, 10), ci_frame=frame, columns=(0, 2))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False, False, True, True, False, False, False, False, False, False],
                                  [False, False, True, True, False, False, False, False, False, False],
                                  [False, False, True, True, False, False, False, False, False, False]])).all()


class TestMaskedSerialTrails:

    def test__pattern_left___mask_only_contains_trails(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_serial_trails_from_ci_frame(shape=(3, 10), ci_frame=frame, columns=(0, 6))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False, False, False, False, True,  True,  True, True,  True,  True],
                                  [False, False, False, False, True,  True,  True, True,  True,  True],
                                  [False, False, False, False, True,  True,  True, True,  True,  True]])).all()
        
    def test__same_as_above_but_uses_invert(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_serial_trails_from_ci_frame(shape=(3, 10), ci_frame=frame, columns=(0, 6), 
                                                                 invert=True)

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[True, True, True, True, False,  False,  False, False,  False,  False],
                                  [True, True, True, True, False,  False,  False, False,  False,  False],
                                  [True, True, True, True, False,  False,  False, False,  False,  False]])).all()

    def test__pattern_left__2_regions__extracts_columns_correctly(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 1, 1, 4), (2, 3, 1, 4)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_left(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_serial_trails_from_ci_frame(shape=(3, 10), ci_frame=frame, columns=(0, 6))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[False, False, False, False, True,  True,  True, True,  True,  True],
                                  [False, False, False, False,  False,  False,  False,  False,  False,  False],
                                  [False, False, False, False, True,  True,  True, True,  True,  True]])).all()
        
    def test__pattern_right___mask_only_contains_trails(self):

        pattern = ci_pattern.CIPatternUniform(normalization=1.0, regions=[(0, 3, 1, 4)])

        frame = ci_frame.ChInj(frame_geometry=ci_frame.QuadGeometryEuclid.bottom_right(), ci_pattern=pattern)

        mask = ci_mask.CIMask.masked_serial_trails_from_ci_frame(shape=(3, 10), ci_frame=frame, columns=(0, 1))

        assert type(mask) == ci_mask.CIMask

        assert (mask == np.array([[True, False, False, False, False, False, False, False, False, False],
                                  [True, False, False, False, False, False, False, False, False, False],
                                  [True, False, False, False, False, False, False, False, False, False]])).all()


import logging

from autofit.tools import pipeline

logger = logging.getLogger(__name__)


class Pipeline(pipeline.Pipeline):
    def run(self, ci_datas, cti_settings, pool=None, data_name=None, assert_optimizer_pickle_matches=True):
        def runner(phase, results):
            return phase.run(ci_datas=ci_datas, cti_settings=cti_settings, results=results, pool=pool)

        return self.run_function(runner, data_name, assert_optimizer_pickle_matches=assert_optimizer_pickle_matches)



import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import itertools

from autocti import exc
from autocti.data.plotters import plotter_util


def plot_array(array, mask=None, extract_array_from_mask=False, as_subplot=False,
               figsize=(7, 7), aspect='equal',
               cmap='jet', norm='linear', norm_min=None, norm_max=None, linthresh=0.05, linscale=0.01,
               cb_ticksize=10, cb_fraction=0.047, cb_pad=0.01, cb_tick_values=None, cb_tick_labels=None,
               title='Array', titlesize=16, xlabelsize=16, ylabelsize=16, xyticksize=16,
               output_path=None, output_format='show', output_filename='array'):
    """Plot an array of hyper as a figure.

    Parameters
    -----------
    array : ndarray or hyper.array.scaled_array.ScaledArray
        The 2D array of hyper which is plotted.
    mask : ndarray of data.mask.Mask
        The masks applied to the hyper, the edge of which is plotted as a set of points over the plotted array.
    extract_array_from_mask : bool
        The plotter array is extracted using the mask, such that masked values are plotted as zeros. This ensures \
        bright features outside the mask do not impact the color map of the plot.
    as_subplot : bool
        Whether the array is plotted as part of a subplot, in which case the grid figure is not opened / closed.
    figsize : (int, int)
        The size of the figure in (rows, columns).
    aspect : str
        The aspect ratio of the hyper, specifically whether it is forced to be square ('equal') or adapts its size to \
        the figure size ('auto').
    cmap : str
        The colormap the array is plotted using, which may be chosen from the standard matplotlib colormaps.
    norm : str
        The normalization of the colormap used to plot the hyper, specifically whether it is linear ('linear'), log \
        ('log') or a symmetric log normalization ('symmetric_log').
    norm_min : float or None
        The minimum array value the colormap map spans (all values below this value are plotted the same color).
    norm_max : float or None
        The maximum array value the colormap map spans (all values above this value are plotted the same color).
    linthresh : float
        For the 'symmetric_log' colormap normalization ,this specifies the range of values within which the colormap \
        is linear.
    linscale : float
        For the 'symmetric_log' colormap normalization, this allowws the linear range set by linthresh to be stretched \
        relative to the logarithmic range.
    cb_ticksize : int
        The size of the tick labels on the colorbar.
    cb_fraction : float
        The fraction of the figure that the colorbar takes up, which resizes the colorbar relative to the figure.
    cb_pad : float
        Pads the color bar in the figure, which resizes the colorbar relative to the figure.
    xlabelsize : int
        The fontsize of the x axes label.
    ylabelsize : int
        The fontsize of the y axes label.
    xyticksize : int
        The font size of the x and y ticks on the figure axes.
    output_path : str
        The path on the hard-disk where the figure is output.
    output_filename : str
        The filename of the figure that is output.
    output_format : str
        The format the figue is output:
        'show' - display on computer screen.
        'png' - output to hard-disk as a png.
        'fits' - output to hard-disk as a fits file.'
    """

    if array is None:
        return

    if extract_array_from_mask and mask is not None:
        array = np.add(array, 0.0, out=np.zeros_like(array), where=np.asarray(mask) == 0)

    plot_figure(array=array, as_subplot=as_subplot,
                figsize=figsize, aspect=aspect, cmap=cmap, norm=norm,
                norm_min=norm_min, norm_max=norm_max, linthresh=linthresh, linscale=linscale)

    plotter_util.set_title(title=title, titlesize=titlesize)
    set_xy_labels_and_ticksize(xlabelsize=xlabelsize, ylabelsize=ylabelsize, xyticksize=xyticksize)

    set_colorbar(cb_ticksize=cb_ticksize, cb_fraction=cb_fraction, cb_pad=cb_pad,
                 cb_tick_values=cb_tick_values, cb_tick_labels=cb_tick_labels)
    plotter_util.output_figure(array, as_subplot=as_subplot, output_path=output_path, output_filename=output_filename,
                               output_format=output_format)
    plotter_util.close_figure(as_subplot=as_subplot)


def plot_figure(array, as_subplot, figsize, aspect, cmap, norm, norm_min, norm_max,
                linthresh, linscale):
    """Open a matplotlib figure and plot the array of hyper on it.

    Parameters
    -----------
    array : ndarray or hyper.array.scaled_array.ScaledArray
        The 2D array of hyper which is plotted.
    as_subplot : bool
        Whether the array is plotted as part of a subplot, in which case the grid figure is not opened / closed.
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float or None
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    figsize : (int, int)
        The size of the figure in (rows, columns).
    aspect : str
        The aspect ratio of the hyper, specifically whether it is forced to be square ('equal') or adapts its size to \
        the figure size ('auto').
    cmap : str
        The colormap the array is plotted using, which may be chosen from the standard matplotlib colormaps.
    norm : str
        The normalization of the colormap used to plot the hyper, specifically whether it is linear ('linear'), log \
        ('log') or a symmetric log normalization ('symmetric_log').
    norm_min : float or None
        The minimum array value the colormap map spans (all values below this value are plotted the same color).
    norm_max : float or None
        The maximum array value the colormap map spans (all values above this value are plotted the same color).
    linthresh : float
        For the 'symmetric_log' colormap normalization ,this specifies the range of values within which the colormap \
        is linear.
    linscale : float
        For the 'symmetric_log' colormap normalization, this allowws the linear range set by linthresh to be stretched \
        relative to the logarithmic range.
    xticks_manual :  [] or None
        If input, the xticks do not use the array's default xticks but instead overwrite them as these values.
    yticks_manual :  [] or None
        If input, the yticks do not use the array's default yticks but instead overwrite them as these values.
    """

    plotter_util.setup_figure(figsize=figsize, as_subplot=as_subplot)

    norm_min, norm_max = get_normalization_min_max(array=array, norm_min=norm_min, norm_max=norm_max)
    norm_scale = get_normalization_scale(norm=norm, norm_min=norm_min, norm_max=norm_max,
                                         linthresh=linthresh, linscale=linscale)

    extent = get_extent(array=array)

    plt.imshow(array, aspect=aspect, cmap=cmap, norm=norm_scale, extent=extent)


def get_extent(array):
    """Get the extent of the dimensions of the array in the units of the figure (e.g. arc-seconds or kpc).

    This is used to set the extent of the array and thus the y / x axis limits.

    Parameters
    -----------
    array : ndarray or hyper.array.scaled_array.ScaledArray
        The 2D array of hyper which is plotted.
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    xticks_manual :  [] or None
        If input, the xticks do not use the array's default xticks but instead overwrite them as these values.
    yticks_manual :  [] or None
        If input, the yticks do not use the array's default yticks but instead overwrite them as these values.
    """
    return np.asarray([0, array.shape[1], 0, array.shape[0]])


def get_normalization_min_max(array, norm_min, norm_max):
    """Get the minimum and maximum of the normalization of the array, which sets the lower and upper limits of the \
    colormap.

    If norm_min / norm_max are not supplied, the minimum / maximum values of the array of hyper are used.

    Parameters
    -----------
    array : ndarray or hyper.array.scaled_array.ScaledArray
        The 2D array of hyper which is plotted.
    norm_min : float or None
        The minimum array value the colormap map spans (all values below this value are plotted the same color).
    norm_max : float or None
        The maximum array value the colormap map spans (all values above this value are plotted the same color).
    """
    if norm_min is None:
        norm_min = array.min()
    if norm_max is None:
        norm_max = array.max()

    return norm_min, norm_max


def get_normalization_scale(norm, norm_min, norm_max, linthresh, linscale):
    """Get the normalization scale of the colormap. This will be scaled based on the input min / max normalization \
    values.

    For a 'symmetric_log' colormap, linthesh and linscale also change the colormap.

    If norm_min / norm_max are not supplied, the minimum / maximum values of the array of hyper are used.

    Parameters
    -----------
    array : ndarray or hyper.array.scaled_array.ScaledArray
        The 2D array of hyper which is plotted.
    norm_min : float or None
        The minimum array value the colormap map spans (all values below this value are plotted the same color).
    norm_max : float or None
        The maximum array value the colormap map spans (all values above this value are plotted the same color).
    linthresh : float
        For the 'symmetric_log' colormap normalization ,this specifies the range of values within which the colormap \
        is linear.
    linscale : float
        For the 'symmetric_log' colormap normalization, this allowws the linear range set by linthresh to be stretched \
        relative to the logarithmic range.
    """
    if norm is 'linear':
        return colors.Normalize(vmin=norm_min, vmax=norm_max)
    elif norm is 'log':
        if norm_min == 0.0:
            norm_min = 1.e-4
        return colors.LogNorm(vmin=norm_min, vmax=norm_max)
    elif norm is 'symmetric_log':
        return colors.SymLogNorm(linthresh=linthresh, linscale=linscale, vmin=norm_min, vmax=norm_max)
    else:
        raise exc.PlottingException('The normalization (norm) supplied to the plotter is not a valid string (must be '
                                    'linear | log | symmetric_log')


def set_xy_labels_and_ticksize(xlabelsize, ylabelsize, xyticksize):
    """Set the x and y labels of the figure, and set the fontsize of those labels.

    The x and y labels are always the distance scales, thus the labels are either arc-seconds or kpc and depend on the \
    units the figure is plotted in.

    Parameters
    -----------
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    xlabelsize : int
        The fontsize of the x axes label.
    ylabelsize : int
        The fontsize of the y axes label.
    xyticksize : int
        The font size of the x and y ticks on the figure axes.
    """
    plt.xlabel('x (pixels)', fontsize=xlabelsize)
    plt.ylabel('y (pixels)', fontsize=ylabelsize)
    plt.tick_params(labelsize=xyticksize)


def set_colorbar(cb_ticksize, cb_fraction, cb_pad, cb_tick_values, cb_tick_labels):
    """Setup the colorbar of the figure, specifically its ticksize and the size is appears relative to the figure.

    Parameters
    -----------
    cb_ticksize : int
        The size of the tick labels on the colorbar.
    cb_fraction : float
        The fraction of the figure that the colorbar takes up, which resizes the colorbar relative to the figure.
    cb_pad : float
        Pads the color bar in the figure, which resizes the colorbar relative to the figure.
    cb_tick_values : [float]
        Manually specified values of where the colorbar tick labels appear on the colorbar.
    cb_tick_labels : [float]
        Manually specified labels of the color bar tick labels, which appear where specified by cb_tick_values.
    """

    if cb_tick_values is None and cb_tick_labels is None:
        cb = plt.colorbar(fraction=cb_fraction, pad=cb_pad)
    elif cb_tick_values is not None and cb_tick_labels is not None:
        cb = plt.colorbar(fraction=cb_fraction, pad=cb_pad, ticks=cb_tick_values)
        cb.ax.set_yticklabels(cb_tick_labels)
    else:
        raise exc.PlottingException('Only 1 entry of cb_tick_values or cb_tick_labels was input. You must either supply'
                                    'both the values and labels, or neither.')

    cb.ax.tick_params(labelsize=cb_ticksize)


def convert_grid_units(array, grid_arc_seconds, units, kpc_per_arcsec):
    """Convert the grid from its input units (arc-seconds) to the input unit (e.g. retain arc-seconds) or convert to \
    another set of units (pixels or kilo parsecs).

    Parameters
    -----------
    array : ndarray or hyper.array.scaled_array.ScaledArray
        The 2D array of hyper which is plotted, the shape of which is used for converting the grid to units of pixels.
    grid_arc_seconds : ndarray or hyper.array.grid_stacks.RegularGrid
        The (y,x) coordinates of the grid in arc-seconds, in an array of shape (total_coordinates, 2).
    units : str
        The units of the y / x axis of the plots, in arc-seconds ('arcsec') or kiloparsecs ('kpc').
    kpc_per_arcsec : float
        The conversion factor between arc-seconds and kiloparsecs, required to plot the units in kpc.
    """
    if units is 'pixels':
        return array.grid_arc_seconds_to_grid_pixels(grid_arc_seconds=grid_arc_seconds)
    elif units is 'arcsec' or kpc_per_arcsec is None:
        return grid_arc_seconds
    elif units is 'kpc':
        return grid_arc_seconds * kpc_per_arcsec


import builtins

import pytest
from autofit.mapper import model_mapper as mm
from autofit.optimize import non_linear as nl

from autocti.pipeline import pipeline as pl


class MockFile(object):
    def __init__(self):
        self.text = None
        self.filename = None

    def write(self, text):
        self.text = text

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass


@pytest.fixture(name="mock_files", autouse=True)
def make_mock_file(monkeypatch):
    files = []

    def mock_open(filename, flag):
        assert flag in ("w+", "w+b")
        file = MockFile()
        file.filename = filename
        files.append(file)
        return file

    monkeypatch.setattr(builtins, 'open', mock_open)
    return files


class Optimizer(object):
    def __init__(self, phase_name):
        self.phase_name = phase_name


class DummyPhase(object):
    def __init__(self, phase_name, phase_tag=None):
        self.masked_image = None
        self.results = None
        self.phase_name = phase_name
        self.phase_tag = phase_tag
        self.phase_path = phase_name

        self.optimizer = Optimizer(phase_name)

    def run(self, ci_datas, cti_settings, results, pool):
        self.ci_datas = ci_datas
        self.cti_settings = cti_settings
        self.results = results
        self.pool = pool
        return nl.Result(mm.ModelInstance(), 1)


class TestPipeline(object):
    def test_run_pipeline(self):
        phase_1 = DummyPhase(phase_name='dummy1')
        phase_2 = DummyPhase(phase_name='dummy2')
        pipeline = pl.Pipeline('', phase_1, phase_2)

        pipeline.run(ci_datas=None, cti_settings=None, pool=None)

        assert len(phase_1.results) == 2
        assert len(phase_2.results) == 2

    def test_addition(self):
        phase_1 = DummyPhase(phase_name='dummy1')
        phase_2 = DummyPhase(phase_name='dummy2')
        phase_3 = DummyPhase(phase_name='dumy3')

        pipeline1 = pl.Pipeline('', phase_1, phase_2)
        pipeline2 = pl.Pipeline('', phase_3)

        assert (phase_1, phase_2, phase_3) == (pipeline1 + pipeline2).phases
